/*************************************************************************
 **                         cncreteMed.c++                              **
 *************************************************************************
 ** the Concrete Mediator for the Interfaces. it's parent class is      **
 ** the abstract vwMediator class.  cncreteMed manages an Observer      **
 ** Pattern where the widgets are Subjects or both Subjects and         **
 ** Observers, other observers would be cncreteeObserv and possibly     **
 ** other Observer widgets (e.g., vwRadioButton)                        **
 ** Author: Fred Dech                                                   **
 ** (C) 2000 Virtual Reality in Medicine Laboratory                     **
 **          University of Illinois at Chicago                          **
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <pfcave.h>
#include <vwSlider.h>
#include <vwFrame.h>
#include <vwValueFrame.h>

#include "cncreteMed.h"
#include "myRadioButton.h"
#include "myPushButton.h"

// static hash_map from myEnums.h
// used in cncreteMed::extractBIdent()
buttonIdentTypeScope::myMap buttonIdentTypeScope::btnMap;

/*
 * NOTE:  there are several static vwWidget member functions called directly
 * by cncreteMed.
 *   static int vwWidget::handleWidgets(pfGroup *, vwWidget * [],
 *                                                 CAVEID, int);
 *   static int vwWidget::checkForClick(int, vwWidget * [], int);
 *   static wdgStruct * vwWidget::parseWdgConfig(char *, char **, char *,
 *                                               pfDCS **, int *);
 *   static void vwWidget::clean_wdgStruct(vwWidget::wdgStruct *);
 * these are helper functions and are pretty generic, it is highly recommended
 * that they be used, but not absolutely necessary.  if you prefer handling
 * this stuff your way, there is nothing inherently problematic with doing
 * so.
 */
cncreteMed::cncreteMed(pfScene * scene, const char *configFile)
	: _scene(scene), _wCount(0),
	  _state(mainInterfaceStateScope::PARENT),
	  _interfaceType(mainInterfaceTypeScope::MAIN_MENU),
	  _ifaceGroup(NULL), _font(NULL), _childMgr(NULL),
	  _nuChildIndex(0), _lastiDCS(-1), _myObserv(NULL)
{
	// instantiate child manager singleton.
	// more than one childMgr is not cool.
	_childMgr = childMgr::instance();

	// instantiate the external interface observer
	_myObserv = myObserv::instance(this);

	// set pfFilePath for loading pfFont in constructor
	pfFilePath(".:../GUI/FONT");
	char path[256] = "../GUI/FONT\0";

	// load the font that'll get used by any widgets that want it.
	if (pfFindFile("Mistr.of", path, R_OK)) {
		_font = pfdLoadFont_type1("Mistr", PFDFONT_EXTRUDED);
	} else {
		fprintf(stderr, "\nWARNING: couldn't load pfFont\n\n");
	}

	// initialize
	for (int nn=0; nn<MAX_INTERFACES; nn++) {
		_iDCS[nn] = NULL;
		for (int oo=0; oo<MAX_ELEMENTS; oo++)
			_wdgList[oo + nn * MAX_ELEMENTS] = NULL;
	}
	// pfGroup that all interface pfDCSs are children of
	_ifaceGroup = new pfGroup;
	_scene->addChild(_ifaceGroup);

	// fill up associative array of strings and enums for buttons
	buttonIdentTypeScope::fillBtnMap();


	// create the widgets described in configFile
	if (createWidgets(const_cast<char *>(configFile)) == NULL)
		fprintf(stderr, "BY THE WAY ... NO GUI WAS LOADED!\n");
}

// destructor.  get rid of all the Performer nodes.
cncreteMed::~cncreteMed()
{
    _iDCS[0]->removeChild(_lite);
    pfDelete(_lite);

    for (int nn=_childMgr->getChildDepth()-1; nn >= 0; nn--)
    {
        cleanUp(nn);
        _childMgr->removeChild();
    }

    for (int nn=0; nn<_wCount; nn++) {
        if (_wdgList[nn]->getDCS()) {
            _iDCS[0]->removeChild(_wdgList[nn]->getDCS());
        } else {
            _iDCS[0]->removeChild(_wdgList[nn]->getSCS());
        }
        delete _wdgList[nn];
    }
    delete _childMgr;

    _ifaceGroup->removeChild(_iDCS[0]);
    _scene->removeChild(_ifaceGroup);

    pfDelete(_iDCS[0]);
    pfDelete(_ifaceGroup);
}

// used by cncreteMed::createWidgets() for specific vwMainButton types
// that are used in this example concrete mediator.
buttonIdentType
cncreteMed::extractBIdent(char *str)
{
    buttonIdentType bIdent = buttonIdentTypeScope::UNKNOWN;

    // string is key in hash_map, use it to set value of bIdent
    // myMap, btnMap declared/defined in myEnums.h
    buttonIdentTypeScope::myMap::iterator itr =
    buttonIdentTypeScope::btnMap.find(const_cast<const char *>(str));

    if (itr != buttonIdentTypeScope::btnMap.end())
        // if we find the hash entry, extract the buttonIdentType.
        bIdent = (*itr).second;

    return bIdent;
}

/*
 * vwMediator::createWidgets() is a pure vitual function and thus must be
 * declared in the concrete mediator's header file, and should be defined
 * in the concrete mediator's source file (here).

 * Load and construct Main Menu interface widgets or other tools described
 * in configFile.
 * some or all child config files are also parsed by this method.
 * in this example, only one of the buttons
 * created in the main menu has a child.  that child points to a config-
 * file for a basic slider interface.
 */
int
cncreteMed::createWidgets(char *configFile)
{
	char * myGuiName = NULL;
	char mypfPath[128] = "../GUI";

	pfDCS * myGuiDCS = NULL;
    // vwWidget::wdgStruct is a one-way linking data class nested inside
    // vwWidget.
	vwWidget::wdgStruct * myWS = NULL;

	int total = 0;

	/* parse configFile and return vwWidget::wdgStruct * linked-list
	 * of widget-info, allocates and initializes *myGuiDCS as parent DCS
     * to all objects in this interface. myGuiName will contain header
     * ID string from configFile.  total is passed by address. it's value
	 * corresponds to the number of structs in the linked list
	 * (number of widget-infos parsed from configFile).
	 */
	myWS = vwWidget::parseWdgConfig(configFile,&myGuiName,
		                            mypfPath,&myGuiDCS,&total);

	// this example has two possible interfaces.  the MAIN_MENU interface
	// and a child of one of the buttons that is a SLIDER interface.

	int realTot = 0;

	_lastiDCS++;  // increment DCS counter (init value is -1)

	// main interface config header
	if (!strncmp("MAIN_MENU", myGuiName,9)) {
		_iDCS[_lastiDCS] = myGuiDCS;
		_ifaceGroup->addChild(_iDCS[_lastiDCS]);

		// create primary interface
		if (!(realTot = createMainInterface(myWS, total))) {
			fprintf(stderr, "cncreteMed::createWidgets(): problem ");
			fprintf(stderr, "creating %s\n",myGuiName);
			return 0;
		}

	// slider interface config header
	} else if (!strncmp("SLIDER", myGuiName,6)) {
		_iDCS[_lastiDCS] = myGuiDCS;
		_ifaceGroup->addChild(_iDCS[_lastiDCS]);

		// set child's start index in childMgr
		_childMgr->setChildStartIndex(_nuChildIndex, _wCount);

		// create child slider interface
		if (!(realTot = createSliderInterface(myWS, total))) {
			fprintf(stderr, "cncreteMed::createWidgets(): problem ");
			fprintf(stderr, "creating %s\n",myGuiName);
			return 0;
		}
		_childMgr->setNumWidgets(_nuChildIndex, realTot);

	} else {
		fprintf(stderr, "cncreteMed::createWidgets(): ");
		fprintf(stderr, "invalid header value of %s\n", myGuiName);
		return 0;
	}

	// free parsed widget class
	delete myWS;

	return 1;
}

int
cncreteMed::createMainInterface(vwWidget::wdgStruct *myWS, int total)
{
	_interfaceType = mainInterfaceTypeScope::MAIN_MENU;

	// make a light for GUI.  this should probably be rewritten using
	// a local light of some sort. Performer does not isolate the effects
	// of a light source to it's branch in the scene-graph as one
	// would suspect it should.
	// will probably add an emissive component to vwWidget so you can set
	// emissive, if you so desire.
	_lite = new pfLightSource;

	_lite->setAmbient(0.45f, 0.45f, 0.45f);
	_lite->setColor(PFLT_DIFFUSE, 1.0f, 1.0f, 1.0f);
	_lite->setPos(0.0f, -5.0f, 0.0f, 1.0f);

	_iDCS[_lastiDCS]->addChild(_lite);

	// extract widget information from linked-list and create the
	// widgets.
	int nn = 0;
	myRadioButton * mb = NULL;
	myPushButton * pb = NULL;

	buttonIdentType mbit;

	while (nn < total) {
		// all widgets in MAIN_MENU in this example myRadioButton and
		// one myPushButton (RESET)
		if (myWS->type == vwWidget::wButton) {
			if ((mbit = extractBIdent(myWS->function)) !=
			             buttonIdentTypeScope::RESET)
			{
			// if it ain't the RESET button
				// instantiate a myRadioButton. 
				mb = new myRadioButton(this, (pfDCS *)myWS->node);
				// set the mainButton's specific function.
				mb->setIdent(mbit);
				// add button to widget list
				_wdgList[nn] = mb;
				// all vwButton's are DCS, add them to interface DCS.
				_iDCS[_lastiDCS]->addChild(mb->getDCS());

				/* put child-list strings into button widget */
				int cn = 0; char  *chl;
				// check if the button has a child
				while (myWS->child_count > 0) {
					chl = *((myWS->child_list) + cn);
					mb->addChild(chl);
					myWS->child_count--;
					cn++;
				}
				mb->setLabelText(myWS, _font);
			} else {
			// must be the RESET button
				// make pushbutton RESET
				pb = new myPushButton(this, (pfDCS *)myWS->node);
				pb->setIdent(mbit);
				_wdgList[nn] = pb;
				_iDCS[_lastiDCS]->addChild(pb->getDCS());
				pb->setLabelText(myWS, _font);
				pb->attach(_myObserv);

			}
		} else {
			fprintf(stderr, "cncreteMed::createWidgets(): ");
			fprintf(stderr, "don't know this widget type.\n");
			total--;
			if (nn > 0 && nn < total) nn--;
			if (myWS->node) pfDelete(myWS->node);
		}
		myWS = myWS->next;  // incriment to next linked struct
		nn++;
	}
	_wCount = total;

	/* in this example, the buttons are radiobuttons (subjects/observers)
	 * NOTE: each button DOES NOT attach to itself as an observer
     * all buttons attach to RESET button
	 */
	for (nn=0; nn<_wCount; nn++) {
		for (int oo=0; oo<_wCount; oo++)
			if (nn != oo) {
				myRadioButton * mrb =
				          dynamic_cast<myRadioButton *>(_wdgList[oo]);
				if (mrb) {
					_wdgList[nn]->attach(mrb);
				}
			}
	}

	return total;
}

int
cncreteMed::createSliderInterface(vwWidget::wdgStruct *myWS, int total)
{
	float center = 0.0;        // dummy data value for slider
	float lo = -180.0;		   // dummy data max
	float hi = 180.0; 		   // dummy data min

	// get slider's value from sliderObserv.
	center = _myObserv->getSliderVal();

	_interfaceType = mainInterfaceTypeScope::SLIDER;

	int nn = 0;

	// these are the classes we expect
	vwSlider * slider = NULL;
	vwFrame * frame = NULL;
	vwValueFrame *vframe = NULL;

	while (nn < total) {
		switch (myWS->type) {
		  case vwWidget::wFrame : // slider frame or valueFrame
			if (!strcmp(myWS->function, "VFRAME1")) {
			// its the value frame
			  vframe = new vwValueFrame(this, (pfSCS *)myWS->node, _font);
			  // init vframe as FLOAT displaying 3 decimals
			  vframe->initValue(center, 3);
			  _iDCS[_lastiDCS]->addChild(myWS->node);
			  _wdgList[_wCount+nn] = vframe;
			} else {
			// it's the vwFrame for the slider
			  frame = new vwFrame(this, (pfSCS *)myWS->node);
			  _iDCS[_lastiDCS]->addChild(myWS->node);
			  _wdgList[_wCount+nn] = frame;
			}
			break;
		  case vwWidget::wSlider :
			slider = new vwSlider(this, (pfDCS *)myWS->node);
			// set slider's local-coord AXIS and range of motion
			slider->setGeomRange(myWS->geom_range, myWS->axis);
			// set slider's data min, max and start value
			slider->setDataParms(lo, hi, center);
		    _iDCS[_lastiDCS]->addChild(myWS->node);
			_wdgList[_wCount+nn] = slider;
			break;
		  default :
			fprintf(stderr, "cncreteMed::createWidgets(): ");
			fprintf(stderr, "don't know this widget type.\n");
			total--;
			if (nn > 0 && nn < total) nn--;
			if (myWS->node)
				pfDelete(myWS->node);
		}
		myWS = myWS->next;  // incriment to next linked struct
		nn++;
	}
	_wCount += total;

	// attach vframe to slider so vframe will update() upon slider's
	// drag events.
	if (slider && vframe)
		slider->attach(vframe);
	else {
		fprintf(stderr, "cncreteMed::createSliderInterface(): unable ");
		fprintf(stderr, "to attach vwframe to slider.\n");
	}

	// attach myObserv to slider
	slider->attach(_myObserv);

	return total;
}


// handleMenu() get's called from main application loop.
int
cncreteMed::handleMenu()
{
	int wOver;
	int click;
	static int dragCond = False;

	/* vwWidget::handleWidgets() is a static helper function which
	 * should be called once per frame. it sets vwWidget->isOver()
	 * according to that state-change in any widget contained in the
	 * widget pointer list it is passed.
	 * it returns an index to currently intersected vwWidget.
	 * if there are no intersections of interest, -1 is returned.
	 * if CAVEBUTTON1 is pressed and a widget is being intersected of a type
	 * that has a drag() method (e.g., vwSlider) handleWidgets() continues
	 * to return the intersection at press-time, until CAVEBUTTON1 is
	 * released. arguments are pfGroup, pointer to widget list (** vwWidget),
	 * total number of entries in widget list. and pointer to an int, who's
     * value is modified by vwWidget::handleWidgets().
	 * last argument is either CAVE_WAND or CAVE_WAND_NAV. CAVE_WAND is
	 * the default value if no argument is passed. this determines
	 * what position and vector are used to set the pfSegSet position and
	 * orientation for widget intersection.
	 */
	wOver = vwWidget::handleWidgets(_scene, _wdgList, _wCount, &click);

	if (click && wOver >= 0) {  // strongly implies a button press condition
		vwButton *vrb = dynamic_cast<vwButton *>(_wdgList[wOver]);
		handleButton(vrb, wOver);
	} else if (wOver >= 0) {
		// check for a drag condition
		if (_wdgList[wOver]->isOfType(vwWidget::wSlider)) {
			vwSlider * sldr = (vwSlider *)_wdgList[wOver];
			if (CAVEBUTTON1) { // this is a drag condition
				sldr->drag();
				dragCond = True;
			} else if (dragCond == True) {
			// upon release set dragged state of slider to False.
				sldr->setDragged(False);
				dragCond = False;
			}
		}
	}
	return wOver + 1;
}

void
cncreteMed::handleButton(vwButton *vwb, int index)
{
	if (!vwb)
		return;

	// toggle press event
	vwb->setPressed();

	int tot = _childMgr->getChildDepth();

	// remove anything below level of current widget selected
	for (int nn=tot-1; nn >=0; nn--) {
		if (index < _childMgr->getChildStartIndex(nn)) {
			fprintf(stderr, "Cleaning up Child # %d\n",nn);
			cleanUp(nn);
			_childMgr->removeChild();
		}
	}

	// new press
	if (vwb->getPressed()) {
		// only worry about 1 child menu
		if (vwb->getNumChildren()) {
			vwb->setExpanded();
			// add new child info to the childM
			_nuChildIndex = _childMgr->addChild(index, vwb,
			                (buttonIdentType)vwb->getIdent());

			// create child interface
			createWidgets(vwb->getChild());
		}
	// old press
	} else {
		vwb->setExpanded(False);
	}
}

// cleanup widgets and performer nodes in stale interface
void
cncreteMed::cleanUp(int index)
{
    int start = _childMgr->getChildStartIndex(index);
    int count = _childMgr->getNumWidgets(index);
    int total = start + count;

    // remove this interface from scenegraph and delete it
    for (int nn = start; nn < total; nn++) {
        if (_wdgList[nn]->getSCS() != NULL)
            _iDCS[_lastiDCS]->removeChild(_wdgList[nn]->getSCS());
        else
            _iDCS[_lastiDCS]->removeChild(_wdgList[nn]->getDCS());

        delete _wdgList[nn];
        _wdgList[nn] = NULL;
    }
    _ifaceGroup->removeChild(_iDCS[_lastiDCS]);

    int val = pfDelete(_iDCS[_lastiDCS]);

    _wCount -= count; // set widget-count to new total widgets.

    _iDCS[_lastiDCS--] = NULL;
    _interfaceType = mainInterfaceTypeScope::MAIN_MENU;
}

