/*************************************************************************
 **                       alphaObserv.c++                               **
 *************************************************************************
 ** Observer for alpha sliders                                          **
 *************************************************************************/
#include "alphaObserv.h"

#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pr/pfMaterial.h>
#include <Performer/pfdu.h>

#include <vwSlider.h>
#include <vwButton.h>

#define TRANSP_SWAP
#undef TRANSP_SWAP

////// alphaObserv Singleton Class ///////
alphaObserv * alphaObserv::_instance = NULL;

// constructor
alphaObserv::alphaObserv(vwMediator *m, pfDCS *gdcs)
	: vwObserver(m), _geomdcs(gdcs), _updateEvent(False)
{
	if (!m) {
		fprintf(stderr, "alphaObserv::alphaObserv: Construction ");
		fprintf(stderr, "with NULL vwMediator pointer is not allowed.\n");
		exit(1);
	}

	init();
	loadModel();
}

// update from alpha sliders or RESET
void
alphaObserv::update(vwSubject *s, void *data)
{
	vwSlider * slider;
	vwButton * btn;

	if (!(slider = dynamic_cast<vwSlider *>(s))) {
		// reset to intial state.
		if (btn = dynamic_cast<vwButton *>(s)) {
			if (btn->getIdent() == buttonIdentTypeScope::RESET) {
				reset(); // set to default transparencies
			}
		}
		return;
	}

	float sVal = *(static_cast<float *>(data));

	int ident = slider->getIdent();

	if (sVal != alphaList[ident].currentAlpha) {
		_updateEvent = True;
		alphaList[ident].currentAlpha = sVal;

		alphaMod(&(alphaList[ident]));
		_updateEvent = False;
	}
}

void
alphaObserv::init()
{
	for (int nn=0; nn<buttonIdentMax; nn++)
	{
		alphaList[nn].ident = buttonIdentTypeScope::NA;
		alphaList[nn].currentAlpha = 1.0;  // set all geoms to 1.0 alpha
		alphaList[nn].geom = NULL;
		alphaList[nn].attached = True;
	}
}

void
alphaObserv::reset()
{
	fprintf(stderr, "alphaObserv::reset()\n");
	for (int nn=0; nn<buttonIdentMax; nn++)
	{
		// assume everything was initially completely opaque
		alphaList[nn].currentAlpha = 1.0;
	}
	for (nn=0; nn<buttonIdentMax; nn++)
	{
		if (alphaList[nn].geom)
			alphaMod(&(alphaList[nn]));
	}
}

/*
 * loadModel() is a brute-force approach to loading the orphan geoms and
 * finding the children geoms by looking into their parents (directories).
 * it uses the enum and name list from myDef.h and chugs through it till
 * it finds a directory and then loads whatever it finds inside.
 */
int
alphaObserv::loadModel()
{
	pfNode *node = NULL;
	pfGroup *grp = NULL;

	char cpath[128] = { 0 };

	/* 
	 * get orphan files first (geometry with no parent (directory))
	 * orphans are the only .pfb here.  all others are inside parent
	 * directorys below this geom directory
	 */
	pfFilePath(const_cast<char *>(_ALPHA_OBSERV_MODEL_PATH));
	for (int nn=0; nn< buttonIdentMax; nn++) {
		// _buttonIdentStr is defined in myDefs.h
		sprintf(cpath, "%s.pfb",_buttonIdentStr[nn]);
		if (node = pfdLoadFile(cpath)) {
			grp = new pfGroup;
			grp->addChild(node);
			_geomdcs->addChild(grp);

			alphaList[nn].ident = buttonIdentType(nn);
			alphaList[nn].geom = grp;
		}

    }

	/*
	 * iterate through Parents (Directorys) and load children
	 * for each parent.
	 * note: current implementation only supports 1 level deep
	 * that is, we do not recursively search for nested directories (Parents).
	 */
	char sysStr[128];
	for (int pp=0; pp<buttonIdentMax; pp++) {
		sprintf(cpath, "%s%s/",_ALPHA_OBSERV_MODEL_PATH,_buttonIdentStr[pp]);
		sprintf(sysStr, "ls -f %s\n",cpath);

		// check if this is a directory.  if so, grab the geometries in it.
		if (!system(sysStr) && strcmp(_buttonIdentStr[pp], "")) {
			pfFilePath(cpath);
			/*
			 * load in geoms under a parent (directory)
			 * NOTE:  we just run through entire list.  this causes
			 * pfdLoadFile Warning messeges which some people find
			 * disconcerting.  i kind of like them, though.
			 */
			for (nn=0; nn<buttonIdentMax; nn++) {
				sprintf(cpath, "%s.pfb",_buttonIdentStr[nn]);

				if (node = pfdLoadFile(cpath)) {
					grp = new pfGroup;
					grp->addChild(node);
					_geomdcs->addChild(grp);

					alphaList[nn].ident = buttonIdentType(nn);
					alphaList[nn].geom = grp;
				}
			}
		}

	}

	// set transparencies
	for (nn=0; nn<buttonIdentMax; nn++) {
		if (alphaList[nn].geom)
			alphaMod(&(alphaList[nn]));
	}
	return 0;
}

int
alphaObserv::alphaMod(alphaStrct * strct)
{
	if (strct->geom) {
		float alpha = strct->currentAlpha;

		// remove geom from scenegraph if its alpha becomes 0.0
		if (alpha == 0.0 && strct->attached) {
			_geomdcs->removeChild(strct->geom);
			strct->attached = False;
		}
		// re-attach orphan to scenegraph if its alpha rises above 0.0
		else if (alpha > 0.0 && !strct->attached) {
			_geomdcs->addChild(strct->geom);
			strct->attached = True;
		}

		// modify orphan's transparency
		if (strct->attached) {
			int numPfChld = strct->geom->getNumChildren();

			for (int ch=0; ch< numPfChld; ch++)
				changeTransp(strct->geom->getChild(ch), alpha);
		}
	}

	return 1;
}

void
alphaObserv::changeTransp(pfNode *node, float alpha)
{
	if (node->isOfType(pfGroup::getClassType())) {
		pfGroup * group = dynamic_cast<pfGroup *>(node);
		int numPfChld = group->getNumChildren();
		for (int ch=0; ch<numPfChld; ch++)
			changeTransp(group->getChild(ch), alpha);
	}
	if (node->isOfType(pfGeode::getClassType())) {
		pfGeode *geode = dynamic_cast<pfGeode *>(node);
		int numGSets = geode->getNumGSets();

		for (int nn=0; nn<numGSets; nn++) {

			pfGeoSet *gset = geode->getGSet(nn);
			pfGeoState *gstate = gset->getGState();

#ifdef TRANSP_SWAP
			if (_updateEvent)
				gstate->setMode(PFSTATE_TRANSPARENCY, PFTR_HIGH_QUALITY);
			else
				gstate->setMode(PFSTATE_TRANSPARENCY, PFTR_FAST);

				gstate->setMode(PFSTATE_TRANSPARENCY,
				                PFTR_BLEND_ALPHA | PFTR_NO_OCCLUDE);
#endif

			if (alpha < 1.0)
				gstate->setMode(PFSTATE_TRANSPARENCY, PFTR_FAST);

			// modify the alpha value of the pfGeode's pfGeoSet's pfGeoState
			if (gset->getAttrBind(PFGS_COLOR4)==PFGS_OVERALL)
			{
				pfVec4 * colors;

				gset->getAttrLists(PFGS_COLOR4, (void **)&colors, NULL);
				(*colors)[3] = alpha;

			} else {
				pfMaterial * mtl = NULL;
				mtl = (pfMaterial *)gstate->getAttr(PFSTATE_FRONTMTL);
				mtl->setColorMode(PFMTL_FRONT,PFMTL_CMODE_OFF);
				mtl->setAlpha(alpha);
			}

		}
	}
}

