/*************************************************************************
 **                         cncreteMed.c++                              **
 *************************************************************************
 ** the Concrete Mediator for the Interfaces. it's parent class is      **
 ** the abstract vwMediator class.  cncreteMed manages an Observer      **
 ** Pattern where the widgets are Subjects or both Subjects and         **
 ** Observers, other observers would be cncreteeObserv and possibly     **
 ** other Observer widgets (e.g., vwIncrementor)                        **
 ** Author: Fred Dech                                                   **
 ** (C) 2001 Virtual Reality in Medicine Laboratory                     **
 **          University of Illinois at Chicago                          **
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <pfcave.h>
#include <vwTextFrame.h>
#include <vwValueFrame.h>

#include "cncreteMed.h"

#include "myRadioButton.h"
#include "mySlider.h"
#include "myPushButton.h"

/*
 * NOTE:  there are several static vwWidget member functions called directly
 * by cncreteMed.
 *   static int vwWidget::handleWidgets(pfScene *, vwWidget * [], int);
 *   static int vwWidget::checkForClick(int, vwWidget * [], int);
 *   static wdgStruct * vwWidget::parseWdgConfig(char *, char **, char *,
 *                                               pfDCS **, int *);
 * these are helper functions and are pretty generic, it is highly recommended
 * that they be used, but not absolutely necessary.  if you prefer handling
 * this stuff your way, there is nothing inherently problematic with doing
 * so.
 */
cncreteMed::cncreteMed(pfScene * scene, pfDCS * gdcs, const char *configFile)
	: _scene(scene), _wCount(0), _mCount(0),
	  _state(mainInterfaceStateScope::PARENT),
	  _interfaceType(mainInterfaceTypeScope::MAIN_MENU),
	  _alphaObserv(NULL), _font(NULL), _ifaceGroup(NULL),
	  _childMgr(NULL), _nuChildIndex(0),
	  _lastiDCS(-1), _geomdcs(gdcs), _pickedDCS(NULL), _wdcs(NULL)
{
	// instantiate child manager singleton.
	// more than one childMgr is not cool.
	_childMgr = childMgr::instance();
	// instantiate the transparency manager observer singleton
	// more than one alphaObserv is also not cool.
	_alphaObserv = alphaObserv::instance(this, _geomdcs);

	// trackball class
	_manip = new manipulator;

	// set pfFilePath for loading pfFont in constructor
	pfFilePath(".:../GUI/FONT");
	char path[256] = "../GUI/FONT\0";

	// load the font that'll get used by any widgets that want it.
	if (pfFindFile("Mistr.of", path, R_OK)) {
		_font = pfdLoadFont_type1("Helvetica", PFDFONT_EXTRUDED);
	} else {
		fprintf(stderr, "\nWARNING: couldn't load pfFont\n\n");
	}

	// initialize some lists and create trunk of our GUI
	// scenegraph branch.
	init();

	// create the widgets described in the main interface configFile
	if (createWidgets(const_cast<char *>(configFile)) == NULL)
		fprintf(stderr, "BY THE WAY ... NO GUI WAS LOADED!\n");
}

// destructor.  get rid of all the Performer nodes.
cncreteMed::~cncreteMed()
{
	_iDCS[0]->removeChild(_lite);
	pfDelete(_lite);

	for (int nn=_childMgr->getChildDepth()-1; nn >= 0; nn--)
	{
		cleanUp(nn);
		_childMgr->removeChild();
	}

	for (nn=0; nn<_wCount; nn++) {
		if (_wdgList[nn]->getDCS()) {
			_iDCS[0]->removeChild(_wdgList[nn]->getDCS());
		} else {
			_iDCS[0]->removeChild(_wdgList[nn]->getSCS());
		}
		delete _wdgList[nn];
	}
	delete _childMgr;
	delete _manip;

	_ifaceGroup->removeChild(_iDCS[0]);
	_scene->removeChild(_ifaceGroup);
	pfDelete(_iDCS[0]);
	pfDelete(_ifaceGroup);
}

// initialize some stuff.  should be smarter and dinamically allocate
// lists.  later
void
cncreteMed::init()
{
	for (int nn=0; nn<MAX_INTERFACES; nn++) {
		_iDCS[nn] = NULL;

		for (int oo=0; oo<MAX_ELEMENTS; oo++)
			_wdgList[oo + nn * MAX_ELEMENTS] = NULL;

	}
	// pfGroup that all interface pfDCSs are children of
	_ifaceGroup = new pfGroup;
	_scene->addChild(_ifaceGroup);
}

// used by cncreteMed::createWidgets() for specific mainButton types
// that are used in this example concrete mediator.
// _buttonIdentStr[] is declared and defined static in myEnums.h
buttonIdentType
cncreteMed::extractEFunction(char *str)
{
    buttonIdentType efunc = buttonIdentTypeScope::UNKNOWN;

	for (int nn=0; nn<buttonIdentMax; nn++)
	
		if (!strcmp(str, _buttonIdentStr[nn]))
		{
			efunc = buttonIdentType(nn);
			break;
		}

    return efunc;
}

/*
 * vwMediator::createWidgets() is a pure vitual function and thus must be
 * declared in the concrete mediator's header file, and should be defined
 * in the concrete mediator's source file (here).

 * Load and construct Main Menu interface widgets or other tools described
 * in configFile.
 * some or all child config files are also parsed by this method.
 * in this example, most of the button widgets have children.
 * note:  child interfaces are only supported for vwButtons and its
 * subclasses.
 */
int
cncreteMed::createWidgets(char *configFile)
{
	char * myGuiName = NULL;
	static char mypfPath[64] = "../GUI";

	pfDCS * myGuiDCS = NULL;
	// vwWidget::wdgStruct is a one-way linking data class nested inside
	// vwWidget.
	vwWidget::wdgStruct * myWS = NULL;

	int total = 0;

	/* parse configFile and return vwWidget::wdgStruct * linked-list
	 * of widget-info, allocates and initializes *myGuiDCS as parent DCS
     * to all objects in this interface. myGuiName will contain header
     * ID string from configFile.  total is passed by address. it's value
	 * corresponds to the number of structs in the linked list
	 * (number of widget-infos parsed from configFile).
	 */
	myWS = vwWidget::parseWdgConfig(configFile,&myGuiName,
		                            mypfPath,&myGuiDCS,&total);

	// this example has two possible interfaces.  the MAIN_MENU interface
	// and a child of one of the buttons that is a SLIDER interface.
	int realTot = 0;

	_lastiDCS++; // increment DCS pointer

	if (!strncmp("MAIN_MENU", myGuiName,9)) {
		_iDCS[_lastiDCS] = myGuiDCS;
		_ifaceGroup->addChild(_iDCS[_lastiDCS]);

		// create main interface
		if (!(realTot = createMainInterface(myWS, total))) {
			fprintf(stderr, "cncreteMed::createWidgets(): problem ");
			fprintf(stderr, "creating %s\n",myGuiName);
			return 0;
		}
	} else {
		_iDCS[_lastiDCS] = myGuiDCS;
		_ifaceGroup->addChild(_iDCS[_lastiDCS]);

		// set child's start index in childMgr
		_childMgr->setChildStartIndex(_nuChildIndex, _wCount);

		// create child interface
		if (!(realTot = createSubRegionInterface(myWS, total))) {
			fprintf(stderr, "cncreteMed::createWidgets(): problem ");
			fprintf(stderr, "creating %s\n",myGuiName);
			return 0;
		}
		_childMgr->setNumWidgets(_nuChildIndex, realTot);
	}

	// delete vwWidget::wdgStruct
	delete myWS;

	return 1;
}

int
cncreteMed::createMainInterface(vwWidget::wdgStruct *myWS, int total)
{
	_interfaceType = mainInterfaceTypeScope::MAIN_MENU;

	// make a light for GUI.  this should probably be rewritten using
	// a local light of some sort. Performer does not isolate the effects
	// of a light source to it's branch in the scene-graph as one
	// would suspect it should.
	// will probably add an emissive component to vwWidget so you can set
	// emissive, if you so desire.
	_lite = new pfLightSource;

	_lite->setAmbient(0.75f, 0.75f, 0.75f);
	_lite->setColor(PFLT_DIFFUSE, 1.0f, 1.0f, 1.0f);
	_lite->setPos(0.0f, -15.0f, 0.0f, 1.0f);

	_iDCS[_lastiDCS]->addChild(_lite);

	// extract widget information from linked-list and create the
	// widgets.
	int nn = 0;

	myRadioButton * mb = NULL;
	myPushButton *rstb = NULL;
	vwButton * bt;

	int manip, scale;

	// want to place-hold manip and scale widgets
	manip = scale = -1;

	while (nn < total) {
		if (myWS->type == vwWidget::wButton) {
			bt = NULL;
			buttonIdentType mbf = buttonIdentTypeScope::UNKNOWN;

			// if it's not RESET, then it's a radiobutton
			if ((mbf = extractEFunction(myWS->function)) !=
			           buttonIdentTypeScope::UNKNOWN &&
			           mbf != buttonIdentTypeScope::RESET)
			{
				/* instantiate a mainbutton.  constructor requires
				 * pointer to concrete mediator and pointer to pfNode.
				 * in this case (vwButton), all are pfDCS.
				 */
				mb = new myRadioButton(this, (pfDCS *)myWS->node);
				// set the mainButton's specific function.
				mb->setIdent(mbf);
				_wdgList[nn] = mb;

				// add button to widget list
				// all vwButton's are DCS, add them to interface DCS.
				_iDCS[_lastiDCS]->addChild(mb->getDCS());

				/* put child-list strings into button widget.
				 * while this loop accommodates multiple children, this
				 * concrete mediator conceptually assumes one child
				 * maximum per button
				 */
				int cn = 0; char  *chl;
				while (myWS->child_count > 0) {
					chl = *((myWS->child_list) + cn);
					mb->addChild(chl);
					myWS->child_count--;
					cn++;
				}
				if (mbf == buttonIdentTypeScope::MANIP)
					manip = nn;
				if (mbf == buttonIdentTypeScope::SCALE)
					scale = nn;

				bt = static_cast<vwButton *>(mb);

				nn++;
			} else if (mbf == buttonIdentTypeScope::RESET) {
				// we know this guy ain't gonna have any children,
				// so we aren't even gonna waste the time.
				fprintf(stderr, "RESET BUTTON\n");
				rstb = new myPushButton(this, (pfDCS *)myWS->node);
				rstb->setIdent(mbf);
				_wdgList[nn] = rstb;
				// set it's push displacement less than default 1.0
				rstb->setPushDistance(0.35);
				_iDCS[_lastiDCS]->addChild(rstb->getDCS());
				bt = static_cast<vwButton *>(rstb);
				nn++;
			} else {
				total--;
				if (nn > 0 && nn < total) nn--;
				if (myWS->node) pfDelete(myWS->node);
			}
			if (bt) {
				// set up label params & create label
				bt->setLabelSize(0.7);
				bt->setLabelHeight(-0.07);
				bt->setExtrudeSize(0.3);
				bt->setLabelText(myWS, _font);
			}
		} else {
			fprintf(stderr, "cncreteMed::createWidgets(): ");
			fprintf(stderr, "don't know this widget type.\n");
			total--;
			if (nn > 0 && nn < total) nn--;
			if (myWS->node) pfDelete(myWS->node);
		}
		myWS = myWS->next;  // increment to next linked struct
	}
	_wCount = _mCount = total;

	/*
	 * attach radionbuttons to each other.
	 * two seperate sets.  first excludes two manipulation buttons.
	 * both include RESET button (rstb) attach.
	 */
	for (nn=0; nn<_mCount; nn++) {
		myRadioButton * vrb1 =
		         dynamic_cast<myRadioButton *> (_wdgList[nn]);
		if (nn != manip && nn != scale) {
			if (vrb1) {
				for (int oo=0; oo<_mCount; oo++) {
					if (oo != manip && oo != scale) {
						myRadioButton * vrb2 =
					        dynamic_cast<myRadioButton *> (_wdgList[oo]);
						if (vrb2) {
							if (nn != oo)
								vrb1->attach(vrb2);
						}
					}
				}
			}
		}
		// attach ALL the radio-buttons to the RESET (rstb) button.
		if (rstb && vrb1)
			rstb->attach(vrb1);
	}

	// attach scale & manip radios to each other
	if (manip >= 0 && scale >= 0) {
		myRadioButton * vrb1 =
			dynamic_cast<myRadioButton *> (_wdgList[manip]);
		myRadioButton * vrb2 =
			dynamic_cast<myRadioButton *> (_wdgList[scale]);

		vrb1->attach(vrb2);
		vrb2->attach(vrb1);

	}

	// only RESET button generates update in _alphaObserv.
	rstb->attach(_alphaObserv);

	return total;
}


/***********************************************************************
 * createSubRegionInterface(): region of interest Interface           **
 ***********************************************************************/
int
cncreteMed::createSubRegionInterface(vwWidget::wdgStruct *myWS, int total)
{
	// these are the classes we expect
	mySlider * slider = NULL;
	vwFrame * frame = NULL;
	vwTextFrame * tframe = NULL;
	vwValueFrame * vframe = NULL;
	myRadioButton * wrb = NULL;

	buttonIdentType bit;

	// place-holders
	int sl, vf;
	sl = vf = -1;

	float lo = 0.0;
	float hi = 1.0;
	float val = 1.0;
	int nn = 0;

	// create the widgets
	// only the slider(s) attach to _alphaObserv.  they are what changes
	// the transparency of the geometries
	while (nn < total) {
	  switch (myWS->type) {
		case vwWidget::wFrame : // a frame
		  if (!strncmp(myWS->function, "FRAME", 5)) { // generic frame
			frame = new vwFrame(this, (pfSCS *)myWS->node);
			_iDCS[_lastiDCS]->addChild(myWS->node);
			_wdgList[_wCount+nn] = frame;
		  } else if (!strncmp(myWS->function, "TEXT", 4)) { // Text Frame
			tframe = new vwTextFrame(this, (pfSCS *)myWS->node);
			_iDCS[_lastiDCS]->addChild(myWS->node);
			_wdgList[_wCount+nn] = tframe;
			tframe->setLabelSize(1.3);
			tframe->setLabelHeight(-0.1);
			tframe->setLabelText(myWS, _font);
		  } else if (!strncmp(myWS->function, "VALUE", 5)) {
			_iDCS[_lastiDCS]->addChild(myWS->node);
			vframe = new vwValueFrame(this, (pfSCS *)myWS->node, _font);
			vframe->setLabelSize(1.5);
			vframe->setLabelHeight(-0.05);
			_wdgList[_wCount+nn] = vframe;
			vf = nn;
		  } else {
			fprintf(stderr, "UNKNOWN FRAME TYPE\n\n");
			total--; if (nn > 0 && nn < total) nn--;
			if (myWS->node) pfDelete(myWS->node);
		  }
		  break;
		case vwWidget::wSlider :
		  slider = new mySlider(this, (pfDCS *)myWS->node);
		  // set the slider's parent identity
		  bit = _childMgr->getParentIdent(_nuChildIndex);
		  slider->setIdent(bit);
		  slider->setGeomRange(myWS->geom_range, myWS->axis);

		  // get current alpha value for this slider's geometry
		  val = _alphaObserv->getCurrentAlpha(bit);

		  slider->setDataParms(lo, hi, val);

		  /*
		   * slider(s) attach _alphaObserv so when their value changes,
		   * an update is generated in _alphaObserv
		   */
		  slider->attach(_alphaObserv);

		  sl = nn;
		  _wdgList[_wCount+nn] = slider;
		  _iDCS[_lastiDCS]->addChild(myWS->node);
		  break;
		case vwWidget::wButton :
		  bit = buttonIdentTypeScope::UNKNOWN;
		  if ((bit = extractEFunction(myWS->function)) !=
				buttonIdentTypeScope::UNKNOWN)
		  {
			wrb = new myRadioButton(this, (pfDCS *)myWS->node);
			wrb->setIdent(bit);

			// how far button is displaced when pressed.
			wrb->setPushDistance(1.0);
			// add button to widget list
			_wdgList[_wCount+nn] = wrb;
			_iDCS[_lastiDCS]->addChild(wrb->getDCS());

			// child strings
			int cn = 0; char  *chl;
			while (myWS->child_count > 0) {
			  chl = *((myWS->child_list) + cn);
			  wrb->addChild(chl);
			  myWS->child_count--;
			  cn++;
			}
			wrb->setLabelSize(0.5);
			wrb->setLabelHeight(-0.04);
			wrb->setLabelText(myWS, _font);
		  }
		  break;
		default :
		  fprintf(stderr, "cncreteMed::createWidgets(): ");
		  fprintf(stderr, "don't know this widget type.\n");
		  total--;
		  if (nn > 0 && nn < total) nn--;
		  if (myWS->node) pfDelete(myWS->node);
	  }
	  myWS = myWS->next;  // increment to next linked struct
	  nn++;
	}

	/*
	 * attach radionbuttons to each other
	 */
	for (nn=0; nn<total; nn++) {
		myRadioButton * vrb1 =
		         dynamic_cast<myRadioButton *> (_wdgList[_wCount+nn]);
		if (vrb1)
			for (int oo=0; oo<total; oo++) {
				myRadioButton * vrb2 =
			      dynamic_cast<myRadioButton *> (_wdgList[_wCount+oo]);
				if (vrb2) {
					if (nn != oo)
						vrb1->attach(vrb2);
				}
			}
	}
	// set value frames value & attach it to slider
	if (vf >=0 && sl >=0) {
		vframe->initValue(slider->getDataValue(), 3);
		slider->attach(vframe);
	}

	_wCount += total; // total number of widgets in mediator

	return total;
}

void
cncreteMed::geomManip()
{
	switch (_state) {
	  case mainInterfaceStateScope::SCALE :
		runScale();
		break;
	  case mainInterfaceStateScope::MANIP :
		runManip();
		break;
	}
}

void
cncreteMed::runScale()
{
	static float inc = 1.1;
	static float sval = 1.0;

	int change = False;

	if (CAVEBUTTON3) {
		sval *= inc;
		change = True;
	} else if (CAVEBUTTON1) {
		sval /= inc;
		change = True;
	} else if (CAVEBUTTON2) {
		sval = 1.0;
		change = True;
	}
	if (change) {
		_geomdcs->setScale(sval, sval, sval);
	}
}

void
cncreteMed::runManip()
{
	static int newManip = False;

	// check for intersection AND NOT CAVEBUTTON1
	if (!CAVEBUTTON1) {
		if (newManip == False) {
			_pickedDCS = NULL;
			// re-set wand pointer to initial (regular) size
			if (_wdcs) _wdcs->setScale(1.0, 1.0, 1.0);
		}
		newManip = True;
	}

	// manip condition
	if (CAVEBUTTON1) {
		// start manip conitional
		if (newManip) {
			newManip = False;
			_pickedDCS = _geomdcs;

			// initialize volume's TM for wand trackball style.
			_manip->init(_pickedDCS);

			// shrink the wand pointer as a visual cue
			if (_wdcs) _wdcs->setScale(0.5, 0.5, 0.5);
		}

		// trackball manipulate _pickedDCS.
		_manip->link();
	}
}

// handleMenu() get's called from main application loop.
int
cncreteMed::handleMenu()
{
	int wOver = _mCount;
	int click = False;
	static int dragCond = False;

	// returns index to intersected widget in _wdgList, else returns
	// -1.
	wOver = vwWidget::handleWidgets(_scene, _wdgList, _wCount, &click);

	// check buttons.  only widget that can return a click=True
	if (click && wOver >= 0) {
		vwButton * vwb = dynamic_cast<vwButton *>(_wdgList[wOver]);
		if (vwb) {
			handleButton(vwb, wOver);
		}
	}
	// if !click there may be a LOCAL slider event
	else if (wOver >= 0)
	{
		vwSlider * sldr = NULL;
		// check for a drag condition if widget is a slider
		if ((sldr = dynamic_cast<vwSlider *>(_wdgList[wOver])))
		{
			if (CAVEBUTTON1) { // this is a drag condition
				sldr->drag(); dragCond = True;
			} else if (dragCond == True) {
				sldr->setDragged(False);
				dragCond = False;
			}
		}
	}
	return wOver + 1;  // False if no intersection, else True
}

void
cncreteMed::handleButton(vwButton *vwb, int index)
{
	if (!vwb)
		return;

	// toggle press event
	vwb->setPressed();

	buttonIdentType bit = buttonIdentType(vwb->getIdent());

	// exclude scale & manip button from this they are seperate.
	// other's need to clear away any open interfaces in the
	// de-selection process of previous radiobutton.
	if (bit != buttonIdentTypeScope::SCALE &&
	    bit != buttonIdentTypeScope::MANIP)
	{
		// get depth of current child interfaces
		int tot = _childMgr->getChildDepth();

		// remove anything below level of current widget selected
		for (int nn=tot-1; nn >=0; nn--) {
			if (index < _childMgr->getChildStartIndex(nn)) {
				fprintf(stderr, "Cleaning up Child # %d\n",nn);
				cleanUp(nn);
				_childMgr->removeChild();
			}
		}
		// in case _state isn't PARENT on RESET.
		if (bit == buttonIdentTypeScope::RESET)
			_state = mainInterfaceStateScope::PARENT;
	}

	// new press
	if (vwb->getPressed()) {
		// maximum 1 child interface per vwButton
		if (vwb->getNumChildren()) {
			vwb->setExpanded();
			// add new child info to the childM
			_nuChildIndex = _childMgr->addChild(index, vwb,
			                 (buttonIdentType)vwb->getIdent());

			// create the child interface
			createWidgets(vwb->getChild());

		}
		else if (bit == buttonIdentTypeScope::MANIP)
		{
	  		_state = mainInterfaceStateScope::MANIP;
		}
		else if (bit == buttonIdentTypeScope::SCALE)
		{
			_state = mainInterfaceStateScope::SCALE;
		}
	} else {
		vwb->setExpanded(False);
		// reset Wand DCS to normal in case it's still small
		if (_wdcs) _wdcs->setScale(1.0, 1.0, 1.0);
		
		if (bit == buttonIdentTypeScope::SCALE ||
		    bit == buttonIdentTypeScope::MANIP)
		{
			_state = mainInterfaceStateScope::PARENT;
		}
	}

}

// cleanup widgets and performer nodes in stale interface 
void
cncreteMed::cleanUp(int index)
{
	buttonIdentType ident = _childMgr->getParentIdent(index);
	int start = _childMgr->getChildStartIndex(index);
	int count = _childMgr->getNumWidgets(index);
	int total = start + count;

	// remove this interface from scenegraph and delete it
	for (int nn = start; nn < total; nn++) {
		if (_wdgList[nn]->getSCS() != NULL) {
			_iDCS[_lastiDCS]->removeChild(_wdgList[nn]->getSCS());
		} else {
			_iDCS[_lastiDCS]->removeChild(_wdgList[nn]->getDCS());
		}
		delete _wdgList[nn];
		_wdgList[nn] = NULL;
	}
	_ifaceGroup->removeChild(_iDCS[_lastiDCS]);
	pfDelete(_iDCS[_lastiDCS]);

	_wCount -= count; // set widget-count to new total widgets.

	_iDCS[_lastiDCS--] = NULL;
	_interfaceType = mainInterfaceTypeScope::MAIN_MENU;
}

