/*************************************************************************
 **                         cncreteMed.h                                **
 *************************************************************************
 ** concrete representation of the abstract vwMediator class.           **
 ** load and manages GUI objects and their signalling to Observer       **
 ** object types.                                                       **
 ** this mediator operates under the model that there is one Main       **
 ** Menu which stays active always.  Main Menu widgets may spawn up     **
 ** to one Sub Menu per widget.  When a Sub Menu is created, the Main   **
 ** Menu still remains active.  In this example mediator, only one      **
 ** widget in the Main Menu will spawn a Sub Menu.  the Sub Menu will   **
 ** be a generic Slider with a Value Button that displays the slider's  **
 ** value dynamically.                                                  **
 ** Author: Fred Dech                                                   **
 ** (C) 2000 Virtual Reality in Medicine Laboratory                     **
 **          University of Illinois at Chicago                          **
 *************************************************************************/

#ifndef _cncreteMed_h_
#define _cncreteMed_h_

#include <Performer/pf/pfScene.h> 
#include <Performer/pf/pfSCS.h>
#include <Performer/pf/pfDCS.h>
#include <vwMediator.h>
#include <vwButton.h>

#include "manipulator.h"
#include "myDefs.h"
#include "childMgr.h"
#include "alphaObserv.h"

const int MAX_ELEMENTS = 32; // number of widgets a menu may have
// max number of seperate interfaces supported by this mediator
const int MAX_INTERFACES = 6;

const char _cncreteMed_Config_Path[64] = "./etc/gui_1.config";

// Concrete Mediator for this set of Observer Pattern user interfaces.
class cncreteMed : public vwMediator
{
public:
	// constructor
	cncreteMed(pfScene *, pfDCS *, const char * m=_cncreteMed_Config_Path);

	// destructor
	virtual ~cncreteMed();

	// run in default Main Menu mode
	int handleMenu();

	// does trackball & scale of geometry
	void geomManip();

	inline void setWandDCS(pfDCS * wdcs)        { _wdcs = wdcs; }
	inline void setGeomDCS(pfDCS * gdcs)        { _geomdcs = gdcs; }
	inline mainInterfaceState getState()        { return _state; }
	inline void setState(mainInterfaceState s)  { _state = s; }

private:
	void init();

	childMgr * _childMgr;
	int _nuChildIndex;

	// this is our transparency observer
	alphaObserv * _alphaObserv;

	// this function is pure virtual in abstract vwMediator class.
	int createWidgets(char *);

	int createMainInterface(vwWidget::wdgStruct *, int);
	int createSubRegionInterface(vwWidget::wdgStruct *, int);

	// some shared memory problem with doing this in the handleButton()
	// routine when dealing with remote activity.  so it works from the
	// routine that calls handleButton().  we'll do it there.
	inline void colorSync(vwButton *vwb)
	{
		if (!vwb->getPressed()) {
			vwb->setColor(vwWidget::ACTIVE);
			vwb->setOver(False);
		}
	}

	pfScene * _scene;
	pfFont  * _font;
	pfGroup * _ifaceGroup;

	pfDCS   * _geomdcs;  // geometry's parent pfDCS
	pfDCS   * _pickedDCS;  // used during trackball manipulation

	// used for trackball-style geom manip
	manipulator * _manip;

	void runScale();
	void runManip();

	pfDCS * _wdcs;        // our wand dcs

	// mediator state variable. currently either PARENT or CHILD mode.
	// defined in wdgEnums.h
	mainInterfaceState _state;

	// mediator sub-state variable.  what child is this?
	// this concrete mediator uses _interfaceType's value internally
	// in order to determine how to handle the widgets.
	// defined in wdgEnums.h
	mainInterfaceType _interfaceType;

	vwWidget::wType _wtype;

	int _mCount;  // MAIN vwButton count
	int _wCount;  // total widget count

	pfDCS * _iDCS[MAX_INTERFACES];
	int _lastiDCS;

	pfLightSource * _lite; // light source for main menu.

	// pointer list of all widgets created in this concrete mediator
	vwWidget * _wdgList[MAX_ELEMENTS * MAX_INTERFACES];

	// used in interface setup routine.
	// pass it a string and get back corresponding buttonIdentType
	buttonIdentType extractEFunction(char *);

	// saves a parent to currently clicked button and sets GUI status.
	void handleButton(vwButton *, int);

	void cleanUp(int);  // clean child out of scene-graph
};

#endif

