/*************************************************************************
 **                         cncreteMed.c++                              **
 *************************************************************************
 ** the Concrete Mediator for the Interfaces. it's parent class is      **
 ** the abstract vwMediator class.  cncreteMed manages an Observer      **
 ** Pattern where the widgets are Subjects or both Subjects and         **
 ** Observers, other observers would be cncreteeObserv and possibly     **
 ** other Observer widgets (e.g., vwIncrementor)                        **
 ** Author: Fred Dech                                                   **
 ** (C) 2001 Virtual Reality in Medicine Laboratory                     **
 **          University of Illinois at Chicago                          **
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <pfcave.h>
#include <vwTextFrame.h>
#include <vwValueFrame.h>

#include "cncreteMed.h"

#include "myRadioButton.h"
#include "myPushButton.h"
#include "fancyButton.h"
#include "fancyFrame.h"
#include "fancyInc.h"
#include "fancyCSlider.h"
#include "fancyWSlider.h"
#include "happySphere.h"

// static hash_map from myEnums.h
// used in cncreteMed::extractBIdent()
buttonIdentTypeScope::myMap buttonIdentTypeScope::btnMap;

/*
 * NOTE:  there are several static vwWidget member functions called directly
 * by cncreteMed.
 *   static int vwWidget::handleWidgets(pfScene *, vwWidget * [], int);
 *   static int vwWidget::checkForClick(int, vwWidget * [], int);
 *   static wdgStruct * vwWidget::parseWdgConfig(char *, char **, char *,
 *                                               pfDCS **, int *);
 * these are helper functions and are pretty generic, it is highly recommended
 * that they be used, but not absolutely necessary.  if you prefer handling
 * this stuff your way, there is nothing inherently problematic with doing
 * so.
 */
cncreteMed::cncreteMed(pfScene * scene, const char *configFile)
	: _scene(scene), _wCount(0),
	  _state(mainInterfaceStateScope::PARENT),
	  _interfaceType(mainInterfaceTypeScope::MAIN_MENU),
	  _fancyObserv(NULL), _font(NULL), _ifaceGroup(NULL),
	  _childMgr(NULL), _nuChildIndex(0),
	  _lastiDCS(-1), _pickedDCS(NULL), _wdcs(NULL)
{
	// instantiate child manager singleton.
	// more than one childMgr is not cool.
	_childMgr = childMgr::instance();
	// instantiate the fancy observer singleton
	_fancyObserv = fancyObserv::instance(this);

	happySphere * hs = happySphere::instance();

	_geomdcs = hs->getDCS();

	// trackball class
	_manip = new manipulator;

	// set pfFilePath for loading pfFont in constructor
	pfFilePath(".:../GUI/FONT");
	char path[256] = "../GUI/FONT\0";

	// load the font that'll get used by any widgets that want it.
	// this should probably be preformed by a static vwWidget function
	// sometime later this century.
	// the Helvetica Performer font was generated using fromtype1 and meshfont
	if (pfFindFile("Mistr.of", path, R_OK)) {
		_font = pfdLoadFont_type1("Mistr", PFDFONT_EXTRUDED);
	} else {
		fprintf(stderr, "\nWARNING: couldn't load pfFont\n\n");
	}

	// initialize some lists and create trunk of our GUI
	// scenegraph branch.
	init();

	// create the widgets described in the main interface configFile
	if (createWidgets(const_cast<char *>(configFile)) == NULL)
		fprintf(stderr, "BY THE WAY ... NO GUI WAS LOADED!\n");
}

// destructor.  get rid of all the Performer nodes.
cncreteMed::~cncreteMed()
{
	_iDCS[0]->removeChild(_lite);
	pfDelete(_lite);

	for (int nn=_childMgr->getChildDepth()-1; nn >= 0; nn--)
	{
		cleanUp(nn);
		_childMgr->removeChild();
	}

	for (int nn=0; nn<_wCount; nn++) {
		if (_wdgList[nn]->getDCS()) {
			_iDCS[0]->removeChild(_wdgList[nn]->getDCS());
		} else {
			_iDCS[0]->removeChild(_wdgList[nn]->getSCS());
		}
		delete _wdgList[nn];
	}
	delete _childMgr;
	delete _manip;

	_ifaceGroup->removeChild(_iDCS[0]);
	_scene->removeChild(_ifaceGroup);
	pfDelete(_iDCS[0]);
	pfDelete(_ifaceGroup);
}

// initialize some stuff.  should be smarter and dinamically allocate
// lists.  later
void
cncreteMed::init()
{
	for (int nn=0; nn<MAX_INTERFACES; nn++) {
		_iDCS[nn] = NULL;

		for (int oo=0; oo<MAX_ELEMENTS; oo++)
			_wdgList[oo + nn * MAX_ELEMENTS] = NULL;

	}
	// pfGroup that all interface pfDCSs are children of
	_ifaceGroup = new pfGroup;
	_scene->addChild(_ifaceGroup);

	// fill up associative array of strings and enums for buttons
	buttonIdentTypeScope::fillBtnMap();

}

// used by cncreteMed::createWidgets() for button ID'ing
buttonIdentType
cncreteMed::extractBIdent(char *str)
{
    buttonIdentType bIdent = buttonIdentTypeScope::UNKNOWN;

	// string is key in hash_map, use it to set value of bIdent
	// myMap, btnMap declared/defined in myEnums.h
	buttonIdentTypeScope::myMap::iterator itr =
	buttonIdentTypeScope::btnMap.find(const_cast<const char *>(str));

	if (itr != buttonIdentTypeScope::btnMap.end())
		// if we find the hash entry, extract the buttonIdentType.
		bIdent = (*itr).second;

	return bIdent;
}

/*
 * vwMediator::createWidgets() is a pure vitual function and thus must be
 * declared in the concrete mediator's header file, and should be defined
 * in the concrete mediator's source file (here).

 * Load and construct Main Menu interface widgets or other tools described
 * in configFile.
 * some or all child config files are also parsed by this method.
 * in this example, most of the button widgets have children.
 * note:  child interfaces are only supported for vwButtons and its
 * subclasses.
 */
int
cncreteMed::createWidgets(char *configFile)
{
	char * myGuiName = NULL;
	static char mypfPath[64] = "../GUI";

	pfDCS * myGuiDCS = NULL;
	// vwWidget::wdgStruct is a one-way linking data class nested inside
	// vwWidget.
	vwWidget::wdgStruct * myWS = NULL;

	int total = 0;

	/* parse configFile and return vwWidget::wdgStruct * linked-list
	 * of widget-info, allocates and initializes *myGuiDCS as parent DCS
     * to all objects in this interface. myGuiName will contain header
     * ID string from configFile.  total is passed by address. it's value
	 * corresponds to the number of structs in the linked list
	 * (number of widget-infos parsed from configFile).
	 */
	myWS = vwWidget::parseWdgConfig(configFile,&myGuiName,
		                            mypfPath,&myGuiDCS,&total);

	// this example has two possible interfaces.  the MAIN_MENU interface
	// and a child of one of the buttons that is a FANCY SLIDER interface.
	int realTot = 0;

	_lastiDCS++; // increment DCS counter (init value is -1)

	if (!strncmp("MAIN_MENU", myGuiName,9)) {
	// get main interface
		_iDCS[_lastiDCS] = myGuiDCS;
		_ifaceGroup->addChild(_iDCS[_lastiDCS]);

		// create main interface
		if (!(realTot = createMainInterface(myWS, total))) {
			fprintf(stderr, "cncreteMed::createWidgets(): problem ");
			fprintf(stderr, "creating %s\n",myGuiName);
			return 0;
		}
	} else {
	// else get secondary (fancy in this case) interface
		_iDCS[_lastiDCS] = myGuiDCS;
		_ifaceGroup->addChild(_iDCS[_lastiDCS]);

		// set child's start index in childMgr
		_childMgr->setChildStartIndex(_nuChildIndex, _wCount);

		// create child interface
		if (!(realTot = createFancyInterface(myWS, total))) {
			fprintf(stderr, "cncreteMed::createWidgets(): problem ");
			fprintf(stderr, "creating %s\n",myGuiName);
			return 0;
		}
		_childMgr->setNumWidgets(_nuChildIndex, realTot);
	}

	// delete vwWidget::wdgStruct
	delete myWS;

	return 1;
}

int
cncreteMed::createMainInterface(vwWidget::wdgStruct *myWS, int total)
{
	_interfaceType = mainInterfaceTypeScope::MAIN_MENU;

	// make a light for GUI.  this should probably be rewritten using
	// a local light of some sort. Performer does not isolate the effects
	// of a light source to it's branch in the scene-graph as one
	// would suspect it should.
	// will probably add an emissive component to vwWidget so you can set
	// emissive, if you so desire.
	_lite = new pfLightSource;

	_lite->setAmbient(0.75f, 0.75f, 0.75f);
	_lite->setColor(PFLT_DIFFUSE, 1.0f, 1.0f, 1.0f);
	_lite->setPos(0.0f, -15.0f, 0.0f, 1.0f);

	_iDCS[_lastiDCS]->addChild(_lite);

	// extract widget information from linked-list and create the
	// widgets.
	int nn = 0;

	myRadioButton * mb = NULL;
	myPushButton *rstb = NULL;
	vwButton * bt;

	int manip, scale;

	// want to place-hold manip and scale widgets
	manip = scale = -1;

	while (nn < total) {
		if (myWS->type == vwWidget::wButton) {
			bt = NULL;
			buttonIdentType mbf;
			// if it's not RESET, then it's a radiobutton
			if ((mbf = extractBIdent(myWS->function)) !=
			           buttonIdentTypeScope::UNKNOWN &&
			           mbf != buttonIdentTypeScope::RESET)
			{
				/* instantiate a mainbutton.  constructor requires
				 * pointer to concrete mediator and pointer to pfNode.
				 * in this case (vwButton), all are pfDCS.
				 */
				mb = new myRadioButton(this, (pfDCS *)myWS->node);
				// set the mainButton's specific function.
				mb->setIdent(mbf);
				_wdgList[nn] = mb;

				// add button to widget list
				// all vwButton's are DCS, add them to interface DCS.
				_iDCS[_lastiDCS]->addChild(mb->getDCS());

				/* put child-list strings into button widget.
				 * while this loop accommodates multiple children, this
				 * concrete mediator conceptually assumes one child
				 * maximum per button
				 */
				int cn = 0; char  *chl;
				while (myWS->child_count > 0) {
					chl = *((myWS->child_list) + cn);
					mb->addChild(chl);
					myWS->child_count--;
					cn++;
				}
				if (mbf == buttonIdentTypeScope::MANIP)
					manip = nn;
				if (mbf == buttonIdentTypeScope::SCALE)
					scale = nn;

				bt = static_cast<vwButton *>(mb);

				nn++;
			} else if (mbf == buttonIdentTypeScope::RESET) {
				// we know this guy ain't gonna have any children,
				rstb = new myPushButton(this, (pfDCS *)myWS->node);
				rstb->setIdent(mbf);
				_wdgList[nn] = rstb;
				// set it's push displacement less than default 1.0
				rstb->setPushDistance(0.35);
				_iDCS[_lastiDCS]->addChild(rstb->getDCS());
				bt = static_cast<vwButton *>(rstb);
				nn++;
			} else {
				total--;
				if (nn > 0 && nn < total) nn--;
				if (myWS->node) pfDelete(myWS->node);
			}
			if (bt) {
				// set up label params & create label
				bt->setPushDistance(1.9);
				bt->setLabelText(myWS, _font);
			}
		} else {
			fprintf(stderr, "cncreteMed::createWidgets(): ");
			fprintf(stderr, "don't know this widget type.\n");
			total--;
			if (nn > 0 && nn < total) nn--;
			if (myWS->node) pfDelete(myWS->node);
		}
		myWS = myWS->next;  // increment to next linked struct
	}
	_wCount = total;

	/*
	 * attach radionbuttons to each other.
	 * two seperate sets.  first excludes two manipulation buttons.
	 * both include RESET button (rstb) attach.
	 */
	for (nn=0; nn<_wCount; nn++) {
		myRadioButton * vrb1 =
		         dynamic_cast<myRadioButton *> (_wdgList[nn]);
		if (nn != manip && nn != scale) {
			if (vrb1) {
				for (int oo=0; oo<_wCount; oo++) {
					if (oo != manip && oo != scale) {
						myRadioButton * vrb2 =
					        dynamic_cast<myRadioButton *> (_wdgList[oo]);
						if (vrb2) {
							if (nn != oo)
								vrb1->attach(vrb2);
						}
					}
				}
			}
		}
		// attach ALL the radio-buttons to the RESET (rstb) button.
		if (rstb && vrb1)
			rstb->attach(vrb1);
	}

	// attach scale & manip radios to each other
	if (manip >= 0 && scale >= 0) {
		myRadioButton * vrb1 =
			dynamic_cast<myRadioButton *> (_wdgList[manip]);
		myRadioButton * vrb2 =
			dynamic_cast<myRadioButton *> (_wdgList[scale]);

		vrb1->attach(vrb2);
		vrb2->attach(vrb1);

		// decided to set scale button inactive because the fancySlider
		// ends up scaling.
		_wdgList[scale]->setActive(False);
	}

	// only RESET has the power... to generate updates in _fancyObserv.
	// that's in the main menu.  fancy menu everything updates fancyObserv
	rstb->attach(_fancyObserv);

	return total;
}


/***********************************************************************
 * createFancyInterface(): fancy slider interface                      *
 * remember.  we're actually create two seperate yet equal fancy       *
 * interfaces.  thus the 1's and 2's.                                  *
 * the 1's will control something through the fancyObserv.             *
 * the 2's will control something else through the fancyObserv.        *
 * we cound make a fancyObserv1 and 2.  this would decrease the RTTI   *
 * but it would be more work for me and i'm in a hurry.                *
 ***********************************************************************/
int
cncreteMed::createFancyInterface(vwWidget::wdgStruct *myWS, int total)
{
	// these are the classes we expect
	float level1, width1;
	float lo, hi;

	_fancyObserv->getSliderLevelWidth(&level1, &width1);
	_fancyObserv->getDataScaleRange(&lo, &hi);

	// incVal is the resolution at which the incrementors change
	float incVal = _fancyObserv->getIncVal();

	// these are the classes we expect
	fancyCSlider * fcSlider1 = NULL;
	fancyInc * flevInc1 = NULL;
	fancyInc * fwthInc1 = NULL;
	fancyWSlider * fwloSlider1 = NULL;
	fancyWSlider * fwhiSlider1 = NULL;
	fancyButton * fmagButt1 = NULL;
	fancyFrame * aFrame1 = NULL;

	// place holders
	int fcs1, fwls1, fwhs1, fli1, fwi1, fbt1;

	// initialize all to -1
	fcs1 = fwls1 = fwhs1 = fli1 = fwi1 = fbt1 = -1;

	// this parsing of myWS is special purpose to the fancyInterface.
	// if we were expecting numerous child interfaces, we'd create some
	// more hash maps and try to handle this in a more generalized way.

	int nn = 0; int missed;
	// create the widgets for fancy slider interface(s)
	while (nn < total) {
		missed = False;
		switch (myWS->type) {
		  case vwWidget::wFrame : // slider frame
			aFrame1 = new fancyFrame(this, (pfSCS *)myWS->node);
			_wdgList[_wCount+nn] = aFrame1;
			break;
		  case vwWidget::wButton : // slider MAGNIFY button
			if (!strncmp(myWS->function, "MAGBUTT1", 8)) {
				fmagButt1 = new fancyButton(this, (pfDCS *)myWS->node);
				fmagButt1->setIdent(buttonIdentTypeScope::MAG1);
				fmagButt1->setInterface(buttonInterfaceTypeScope::FANCY);
				_wdgList[_wCount+nn] = fmagButt1;
				fmagButt1->setSpacingScale(1.2, 1.0, 1.0);
				fmagButt1->setLabelHeight(-0.06);
				fmagButt1->setLabelText(myWS, _font);
				fmagButt1->setPushDistance(0.22);
				fbt1 = nn;
			} else {
				fprintf(stderr, "cncreteMed::createFancyInterface():  ");
				fprintf(stderr, "Unknown Button Type\n");
				total--; if (nn > 0 && nn < total) nn--; missed = True;
			}
			break;
		  case vwWidget::wSlider :
			// SLIDER GROUP
			if (!strncmp(myWS->function, "SLIDERM1", 8)) {
			// level slider of group 1
				fcSlider1 = new fancyCSlider(this, (pfDCS *)myWS->node);
				fcSlider1->setIdent(sliderIdentTypeScope::SLIDERC1);
				fcSlider1->setGeomRange(myWS->geom_range, myWS->axis);
				fcSlider1->setDataParms(lo, hi, level1);
				fcs1 = nn;
				_wdgList[_wCount+nn] = fcSlider1;
			} else if (!strncmp(myWS->function, "SLIDERL1", 8)) {
			// Low slider of group 1
				fwloSlider1= new fancyWSlider(this, (pfDCS *)myWS->node);
				fwloSlider1->setIdent(sliderIdentTypeScope::SLIDERL1);
				fwls1 = nn;
				_wdgList[_wCount+nn] = fwloSlider1;
			} else if (!strncmp(myWS->function, "SLIDERH1", 8)) {
			// High slider of group 1
				fwhiSlider1= new fancyWSlider(this, (pfDCS *)myWS->node);
				fwhiSlider1->setIdent(sliderIdentTypeScope::SLIDERH1);
				fwhs1 = nn;
				_wdgList[_wCount+nn] = fwhiSlider1;
			} else {
				fprintf(stderr, "cncreteMed::createFancyInterface(): ");
				fprintf(stderr, "Unknown Slider Type\n");
				total--; if (nn > 0 && nn < total) nn--; missed = True;
			}
			break;
		  case vwWidget::wIncrementor :
			// INCREMENTOR GROUP
			if (!strncmp(myWS->function, "INCBUTM1", 8)) {
			// incrementor for center.
				flevInc1 = new fancyInc(this, (pfSCS *)myWS->node,
				                            _font);
				// init as float
				flevInc1->setDataParms(lo, hi, incVal);
				flevInc1->setCenter(level1);
				flevInc1->setInterface(
				         incrementorInterfaceTypeScope::FANCY);
				flevInc1->setIdent(incrementorIdentTypeScope::CENTER1);

				fli1 = nn;
				_wdgList[_wCount+nn] = flevInc1;
			} else if (!strncmp(myWS->function, "INCBUTW1", 8)) {
			// incrementor for width
				fwthInc1 = new fancyInc(this, (pfSCS *)myWS->node,
				                                   _font);
				// width init as float
				fwthInc1->setDataParms(0.0, (hi-lo), incVal);
				fwthInc1->setWidth(width1);
				fwthInc1->setInterface(
				         incrementorInterfaceTypeScope::FANCY);
				fwthInc1->setIdent(incrementorIdentTypeScope::WIDTH1);

				fwi1 = nn;
				_wdgList[_wCount+nn] = fwthInc1;

			} else {
				fprintf(stderr, "cncreteMed::createFancyInterface(): ");
				fprintf(stderr, "Unknown Incrementor Type\n");
				total--; if (nn > 0 && nn < total) nn--; missed = True;
			}
			break;
		  default :
			fprintf(stderr, "cncreteMed::createFancyInterface(): ");
			fprintf(stderr, "don't know this widget type.\n");
			total--;
			if (nn > 0 && nn < total) nn--; missed = True;
		}
		if (!missed)
			_iDCS[_lastiDCS]->addChild(myWS->node);
		else {
			if (myWS->node)
				pfDelete(myWS->node);
		}

		myWS = myWS->next;  // increment to next linked struct
		nn++;
	}
	_wCount += total; // _wCount is total number of widgets in mediator

	/* EVENT MAPPING via attach() and parameter initialization
	 * it isn't as bad as it looks.  just think about it this way:
     * 1. whenever the level slider changes, the two width sliders must move
     *    along with it (update) and the level Inc must update.
	 * 2. same if the level inc is changed.
	 * 3. whenever one of the two width sliders moves, the other must be
	 *    updated and the width inc as well.
	 * 4. if the width inc is changed, the two width sliders must be updated.
	 * 5. the MAG button must update everything except the level slider with
	 *    which it has a special relationship.  the frame also gets updated.
	 *    (it turns red when magnification is active).
	 * 6. most widgets here (not frame) must update the fancyObserv.
	 * do remaining initialization and attach obeservers to subjects.
	 * make sure we got all the sliders and then initialize the width one's
	 * to the level slider.  attach width's to level slider.
	 * if we didn't get the sliders, we may as well not bother with the rest
	 */
	if (fcs1 >= 0 && fwls1 >= 0 && fwhs1 >= 0) {
		// initialize and set width sliders
		fwloSlider1->init(fcSlider1);
		fwhiSlider1->init(fcSlider1);

		// specify hi side and low side sliders denoting width
		fwloSlider1->setSide(sliderIdentTypeScope::SLIDERL1, width1);
		fwhiSlider1->setSide(sliderIdentTypeScope::SLIDERH1, width1);

		// low and high slider are attached to center slider
		fcSlider1->attach(fwloSlider1);
		fcSlider1->attach(fwhiSlider1);
		// and each to the other
		fwloSlider1->attach(fwhiSlider1);
		fwhiSlider1->attach(fwloSlider1);
		// all have the fancyObserv attached to them
		fcSlider1->attach(_fancyObserv);
		fwloSlider1->attach(_fancyObserv);
		fwhiSlider1->attach(_fancyObserv);

		// verify that we got the fancyInc's
		if (fli1 >= 0 && fwi1 >= 0) {
			flevInc1->setDataRange(fcSlider1->getDataRange());
			flevInc1->setGeomRange(fcSlider1->getGeomRange());
			flevInc1->setGeomMin(fcSlider1->getGeomMin());
			flevInc1->setGeomMax(fcSlider1->getGeomMax());
			flevInc1->setLabelSize(1.2);
			flevInc1->setSpacingScale(1.2, 1.0, 1.0);
			flevInc1->initValue("CENTER", level1);
	
			fwthInc1->setDataRange(fcSlider1->getDataRange());
			fwthInc1->setLabelSize(1.2);
			fwthInc1->setSpacingScale(1.2, 1.0, 1.0);
			fwthInc1->initValue("WIDTH", width1);

			// center button is attached to level slider & vise-versa
			fcSlider1->attach(flevInc1);
			flevInc1->attach(fcSlider1);	

			// both low and high sliders are attached to center button
			flevInc1->attach(fwloSlider1);
			flevInc1->attach(fwhiSlider1);
			// both low and high sliders are attached to width button
			fwthInc1->attach(fwloSlider1);
			fwthInc1->attach(fwhiSlider1);

			// width button is attached to low and high sliders
			fwloSlider1->attach(fwthInc1);
			fwhiSlider1->attach(fwthInc1);

			// both incs have fancyObserver attached to them
			flevInc1->attach(_fancyObserv);
			fwthInc1->attach(_fancyObserv);
		}

		// segMag button
		if (fbt1 >= 0) {
			fmagButt1->setCenter(level1);
			fmagButt1->setWidth(width1);
			fmagButt1->setOrigScaleRange(lo, hi);
			fmagButt1->setMagScaleRange(lo, hi);

			// fancyButton get's a pointer to its center slider.
			// it keeps this pointer because other widgets re-initialize
			// their datapars according to the center slider.  therefore
			// center slider must be updated in doProcessPress() before
			// notify().  otherwise there is no guarantee that center slider
			// will be updated first.
			fmagButt1->setCenterSlider(fcSlider1);
			float val = level1;
			// put current MAG values into magButton
			fmagButt1->update(fcSlider1, &val);

			// attach both width sliders and both fancyInc's to MAG
			// fancyButton.  center slider does not need to be attached.
			fmagButt1->attach(fwloSlider1);
			fmagButt1->attach(fwhiSlider1);

			// don't need to check for NULL observers. vwLib does that
			fmagButt1->attach(flevInc1);
			fmagButt1->attach(fwthInc1);
			if (aFrame1)
				fmagButt1->attach(aFrame1);

			// attach the MAG fancyButton to all widgets in interface.
			fcSlider1->attach(fmagButt1);
			fwloSlider1->attach(fmagButt1);
			fwhiSlider1->attach(fmagButt1);

			if (fli1 >= 0 && fwi1 >= 0) {
				flevInc1->attach(fmagButt1);
				fwthInc1->attach(fmagButt1);
			}
		}
	}
	return total;
}


void
cncreteMed::geomManip()
{
	switch (_state) {
	  case mainInterfaceStateScope::SCALE :
		runScale();
		break;
	  case mainInterfaceStateScope::MANIP :
		runManip();
		break;
	}
}

void
cncreteMed::runScale()
{
	static float inc = 1.1;
	static float sval = 1.0;

	int change = False;

	if (CAVEBUTTON3) {
		sval *= inc;
		change = True;
	} else if (CAVEBUTTON1) {
		sval /= inc;
		change = True;
	} else if (CAVEBUTTON2) {
		sval = 1.0;
		change = True;
	}
	if (change) {
		_geomdcs->setScale(sval, sval, sval);
	}
}

void
cncreteMed::runManip()
{
	static int newManip = False;

	// check for intersection AND NOT CAVEBUTTON1
	if (!CAVEBUTTON1) {
		if (newManip == False) {
			_pickedDCS = NULL;
			// re-set wand pointer to initial (regular) size
			if (_wdcs) _wdcs->setScale(1.0, 1.0, 1.0);
		}
		newManip = True;
	}

	// manip condition
	if (CAVEBUTTON1) {
		// start manip conitional
		if (newManip) {
			newManip = False;
			_pickedDCS = _geomdcs;

			// initialize volume's TM for wand trackball style.
			_manip->init(_pickedDCS);

			// shrink the wand pointer as a visual cue
			if (_wdcs) _wdcs->setScale(0.5, 0.5, 0.5);
		}

		// trackball manipulate _pickedDCS.
		_manip->link();
	}
}

// handleMenu() get's called from main application loop.
int
cncreteMed::handleMenu()
{
	int wOver = _wCount;
	int click = False;
	static int dragCond = False;

	// returns index to intersected widget in _wdgList, else returns
	// -1.
	wOver = vwWidget::handleWidgets(_scene, _wdgList, _wCount, &click);

	// check buttons.  only widget that can return a click=True
	if (click && wOver >= 0) {
		vwButton * vwb = dynamic_cast<vwButton *>(_wdgList[wOver]);
		if (vwb) {
			handleButton(vwb, wOver);
		}
	}
	// if !click there may be a LOCAL slider event
	else if (wOver >= 0)
	{
		vwSlider * sldr = NULL;
		fancyInc * fInc = NULL;

		// check for a drag condition if widget is a slider
		if ((sldr = dynamic_cast<vwSlider *>(_wdgList[wOver])))
		{
			if (CAVEBUTTON1) { // this is a drag condition
				sldr->drag(); dragCond = True;
			} else if (dragCond == True) {
				sldr->setDragged(False);
				dragCond = False;
			}
		}
		// else check for increment or decrement
		else if ((fInc = dynamic_cast<fancyInc *>(_wdgList[wOver])))
		{
			if (CAVEButtonChange(3) && !CAVEBUTTON3)
				fInc->increment();
			else if (CAVEButtonChange(1) && !CAVEBUTTON1)
				fInc->decrement();
		}
	}
	return wOver + 1;  // False if no intersection, else True
}

void
cncreteMed::handleButton(vwButton *vwb, int index)
{
	if (!vwb)
		return;

	// toggle press event
	vwb->setPressed();

	buttonIdentType bit = buttonIdentType(vwb->getIdent());

	// exclude scale, manip & mag button from this they are seperate.
	// other's need to clear away any open interfaces in the
	// de-selection process of previous radiobutton.
	if (bit != buttonIdentTypeScope::SCALE &&
	    bit != buttonIdentTypeScope::MANIP &&
		bit != buttonIdentTypeScope::MAG1)
	{
		// get depth of current child interfaces
		int tot = _childMgr->getChildDepth();

		// remove anything below level of current widget selected
		for (int nn=tot-1; nn >=0; nn--) {
			if (index < _childMgr->getChildStartIndex(nn)) {
				fprintf(stderr, "Cleaning up Child # %d\n",nn);
				cleanUp(nn);
				_childMgr->removeChild();
			}
		}
		// in case _state isn't PARENT on RESET.
		if (bit == buttonIdentTypeScope::RESET)
			_state = mainInterfaceStateScope::PARENT;
	}

	// new press
	if (vwb->getPressed()) {
		// maximum 1 child interface per vwButton
		if (vwb->getNumChildren()) {
			vwb->setExpanded();
			// add new child info to the childM
			_nuChildIndex = _childMgr->addChild(index, vwb,
			                 (buttonIdentType)vwb->getIdent());

			// create the child interface
			createWidgets(vwb->getChild());

		}
		else if (bit == buttonIdentTypeScope::MANIP)
		{
	  		_state = mainInterfaceStateScope::MANIP;
		}
		else if (bit == buttonIdentTypeScope::SCALE)
		{
			_state = mainInterfaceStateScope::SCALE;
		}
	} else {
		vwb->setExpanded(False);
		// reset Wand DCS to normal in case it's still small
		if (_wdcs) _wdcs->setScale(1.0, 1.0, 1.0);
		
		if (bit == buttonIdentTypeScope::SCALE ||
		    bit == buttonIdentTypeScope::MANIP)
		{
			_state = mainInterfaceStateScope::PARENT;
		}
	}

}

// cleanup widgets and performer nodes in stale interface 
void
cncreteMed::cleanUp(int index)
{
	int start = _childMgr->getChildStartIndex(index);
	int count = _childMgr->getNumWidgets(index);
	int total = start + count;

	// remove this interface from scenegraph and delete it
	for (int nn = start; nn < total; nn++) {
		if (_wdgList[nn]->getSCS() != NULL)
			_iDCS[_lastiDCS]->removeChild(_wdgList[nn]->getSCS());
		else
			_iDCS[_lastiDCS]->removeChild(_wdgList[nn]->getDCS());

		delete _wdgList[nn];
		_wdgList[nn] = NULL;
	}
	_ifaceGroup->removeChild(_iDCS[_lastiDCS]);

	int val = pfDelete(_iDCS[_lastiDCS]);

	_wCount -= count; // set widget-count to new total widgets.

	_iDCS[_lastiDCS--] = NULL;
	_interfaceType = mainInterfaceTypeScope::MAIN_MENU;
}

