/*************************************************************************
 **                         fancyWSlider.c++                            **
 *************************************************************************
 ** fancyWSlider class works as the window-width upper or lower         **
 ** delimiting sliders for a fancy slider interface.  it receives       **
 ** update events from other fancyWSlider, fancyCSlider, fancyInc,      **
 ** & fancyButton.                                                      **
 ** Author: Fred Dech                                                   **
 ** (C) 1999 Virtual Reality in Medicine Laboratory                     **
 **          University of Illinois at Chicago                          **
 *************************************************************************/

#include "fancyWSlider.h"
#include "fancyInc.h"
#include "fancyButton.h"

fancyWSlider::fancyWSlider(vwMediator *m, pfDCS *dcs)
	: fancyCSlider(m,dcs), _geomCenter(0.),
	  _geomWidth(0.), _dataCenter(0.), _dataWidth(0.)
{

}

// initialize WSlider according to values of CSlider it is associated
// with.
void
fancyWSlider::init(vwSlider *s)
{
	if (!s) {
		fprintf(stderr, "fancyWSlider::init(): passed NULL slider pointer\n");
		return;
	}

	_dataRange = s->getDataRange();
	_dataMin = s->getDataMin();
	_dataMax = s->getDataMax();
	_dataCenter = s->getDataValue();

	_geomRange = s->getGeomRange();
	_geomMin = s->getGeomMin();
	_geomMax = s->getGeomMax();
	_geomCenter = s->getGeomValue();

	_axis = s->getAxis();

	// pass address of _dataWidth as (void * data) arg during updates
	setSubjectData(&_dataWidth);
}

void
fancyWSlider::setSide(int ident, float datawidth)
{
	_dataWidth = datawidth;

	_geomWidth = (_dataWidth * _geomRange) / _dataRange;

	if (ident == sliderIdentTypeScope::SLIDERL1 ||
	    ident == sliderIdentTypeScope::SLIDERL2)
	{
		_geomPos = _geomCenter - (_geomWidth / 2.);
	} else {
		_geomPos = _geomCenter + (_geomWidth / 2.);
	}
	updateDrag(_geomPos);
}

// overloaded to accomodate CAVERN remote client event call on this
// slider class.
void
fancyWSlider::setSide(int ident, float datawidth, int isevent)
{
	setSide(ident, datawidth);
	notify();
}

// overridden vwSlider::applyDrag().  get's called from
// vwSlider::computeDrag().  nvals[] holds transformed geom
// position.  index is index of AXIS that will become _geomPos;
void
fancyWSlider::applyDrag(float nvals[], int index)
{
	int ident = getIdent();

	if (ident == sliderIdentTypeScope::SLIDERL1 ||
	    ident == sliderIdentTypeScope::SLIDERL2)
	{
		if (_geomPos >= _geomCenter)
			_geomPos = _geomCenter;

	} else if (ident == sliderIdentTypeScope::SLIDERH1 ||
	           ident == sliderIdentTypeScope::SLIDERH2)
	{
		if (_geomPos <= _geomCenter)
			_geomPos = _geomCenter;
	}

	// low and high slider will always be symmetrical around _geomCenter;
	_geomWidth = (_geomPos - _geomCenter) * 2.0;
	if (_geomWidth < 0.0) _geomWidth *= -1.0;

	_dataWidth = (_geomWidth * _dataRange) / _geomRange;

	// clip slider if it trys to expand range
	if (_dataWidth > _dataRange) {
		setSide(ident, _dataRange);
	} else {
		nvals[index] = _geomPos;

		// apply new geom position
		vwSlider::_mtx.setRow(3, nvals[0], nvals[1], nvals[2], nvals[3]);
		getDCS()->setMat(vwSlider::_mtx);
	}
}

void
fancyWSlider::update(vwSubject *s, void *data)
{
	float val;
	float width;

	// we care about fancyCSliders, fancyWSliders and fancyInc
	vwSlider * slider = dynamic_cast<vwSlider *>(s);
	fancyInc * inc = dynamic_cast<fancyInc *>(s);
	fancyButton * butt = dynamic_cast<fancyButton *>(s);

	if (slider) {
		// get slider's specific ID
		int ident = slider->getIdent();

		if (ident == sliderIdentTypeScope::SLIDERC1) {
			// adjust _geomCenter
			_geomCenter = slider->getGeomValue();
			_dataCenter = slider->getDataValue();
			width = _dataWidth;
		} else {                 // must be the fancyWSlider
			fancyWSlider *fwSlider = dynamic_cast<fancyWSlider *>(s);
			if (!fwSlider)
				return;
			// get new dataWidth value from fancyWSlider
			width = fwSlider->getDataWidth();
		}
	} else if (inc) {
		if (inc->getIdent() == incrementorIdentTypeScope::WIDTH1) {
			width = inc->getWidth();
		} else {
			val = inc->getCenter();
			_dataCenter = val;
			_geomCenter = inc->getGeomVal();
			width = _dataWidth;
		}
	} else if (butt) {
		// it's the Magnify button, re-initialize to MAGed center slider.
		init(butt->getCenterSlider());
		width = butt->getWidth();
	} else
		return;

	// adjust our position
	setSide(getIdent(), width);
}

